/* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */

#include "ns3/gpsr-module.h"
#include "ns3/vector.h"
#include "ns3/string.h"
#include "ns3/socket.h"
#include "ns3/double.h"
#include "ns3/config.h"
#include "ns3/log.h"
#include "ns3/command-line.h"
#include "ns3/ns2-mobility-helper.h"
#include "ns3/mobility-model.h"
#include "ns3/yans-wifi-helper.h"
#include "ns3/position-allocator.h"
#include "ns3/mobility-helper.h"
#include "ns3/internet-stack-helper.h"
#include "ns3/ipv4-address-helper.h"
#include "ns3/ipv4-interface-container.h"
#include <iostream>

#include "ns3/ocb-wifi-mac.h"
#include "ns3/wifi-80211p-helper.h"
#include "ns3/wave-mac-helper.h"

#include "ns3/network-module.h"
#include "ns3/applications-module.h"
#include "ns3/udp-client.h"
#include "ns3/udp-server.h"

#include "ns3/netanim-module.h"
#include "ns3/flow-monitor-module.h"

#include "ns3/waypoint-mobility-model.h"

#define SIM_DURATION 15
//175
#define POLICE_COUNT 3
#define NODE_COUNT 50
#define SCALE 0.1

using namespace ns3;
using namespace gpsr;

#define ASAP Simulator::Now() + Seconds (0.000000001)

NS_LOG_COMPONENT_DEFINE ("tars-sim");

  /**************************************************************************
	 HANDLE DYNAMIC TOGGLING OF STREAM STATE
  ***************************************************************************/

// video streaming data
struct StreamData
{
  bool streaming;
  bool policeCar;
};

// store video streaming data for each node
std::vector<StreamData*> nodeStreams;

// globally store udp apps
ApplicationContainer serverApps;
ApplicationContainer clientApps;

// handle the updates of nodes' streaming
#define SCALING_DURATION 5

// keep track of the radius size
std::vector<uint64_t> radii;
// store data for graphing
std::vector<int> memberCount;
// checker for whether all police nodes are in region
unsigned int inZone =  (1 << POLICE_COUNT) - 1;
void videoStreamingHandler(int policeCount)
{
  static int secs;
  static int scaling = 0;
  bool scaleUp = true;

  //output tracking
  if (secs % 5 == 0)
    std::cout << ">>>> " << secs << std::endl;
  secs++;

  NodeContainer nc = NodeContainer::GetGlobal ();
  // store coords of members in cluster
  std::vector<std::pair<uint64_t, uint64_t> > members;
  int currentRadius;

  int j = 0;
  // count the number of members in the zone
  memberCount.push_back (0);
  for (NodeContainer::Iterator i = nc.Begin (); i != nc.End (); ++i, ++j)
    {
      Ptr<Node> n = (*i);
      Ptr<Ipv4> ipv4 = n->GetObject<Ipv4>();
      Ptr<Ipv4RoutingProtocol> p = ipv4->GetRoutingProtocol ();
      Ptr<ns3::gpsr::RoutingProtocol> rp = DynamicCast<ns3::gpsr::RoutingProtocol> (p);

      // add the coords of nodes in cluster
      if (rp->InCluster ())
        {
          std::pair<uint64_t, uint64_t> xy = {rp->GetX (), rp->GetY ()};
          members.push_back (xy);
          currentRadius = rp->GetRadius ();

          // add member to count
          memberCount.back()++;
        }

      // skip police cars, they will always receive
      if (nodeStreams[j]->policeCar)
        {
          //if (rp->InCluster ())
          //  inZone &= ~(1 << j);
          continue;
        }
      //if (!inZone)
        {
          //Simulator::Stop ();
        }

      // update radii store
      if (j == policeCount)
        radii.push_back (rp->GetRadius ());

      // reduce time until next permitted scaling
      if (scaling)
        scaling--;

      if (rp->GetBufferRemaining () < 62)
        {
          // don't want to scale up
          scaleUp = false;
          //std::cout << "Need to scale: " << rp->GetBufferRemaining () <<
          //  std::endl;
        }

      if (rp->ShouldBeStreaming ())
        {
          // not streaming, start
          if (!nodeStreams[j]->streaming)
            {
              std::cout << "commence streaming...\n";

              // go through the clients and launch them
              for (uint32_t c = policeCount * j; c < (uint32_t) nodeStreams.size () &&
                     c < (uint32_t) policeCount * j + policeCount; ++c)
                {
                  Ptr<Application> a = clientApps.Get (c);
                  Ptr<UdpClient> cl = DynamicCast<UdpClient> (a);
                  cl->SetStartTime (ASAP);

                }
              // toggle state tracker
              nodeStreams[j]->streaming = true;
            }
        }
      else if (nodeStreams[j]->streaming)
        {
          // shouldn't be streaming so
          // node should cease
          std::cout << "cease streaming\n";

          // go through the clients and switch them off
          for (uint32_t c = policeCount * j; c < (uint32_t) nodeStreams.size () &&
                 c < (uint32_t) policeCount * j + policeCount; ++c)
            {
              Ptr<Application> a = clientApps.Get (c);
              Ptr<UdpClient> cl = DynamicCast<UdpClient> (a);
              cl->SetStopTime (ASAP);
            }
          // toggle state tracker
          nodeStreams[j]->streaming = false;
        }
    }

  // check we can scale and we found a node in the cluster zone
  if (scaling || members.size () <= 2)
    return;

  //std::cout << "scaling..." << std::endl;

  // calculate new radius
  int newRadius;
  if (scaleUp)
    newRadius = RoutingProtocol::ScaleZoneUp(currentRadius, members);
  else
    newRadius = RoutingProtocol::ScaleZoneDown(currentRadius, members);
  // don't want a zero radius
  if (!newRadius)
    return;

  // scale the cluster to the new radius
  for (NodeContainer::Iterator i = nc.Begin (); i != nc.End (); ++i)
    {
      Ptr<Node> n = (*i);
      Ptr<Ipv4> ipv4 = n->GetObject<Ipv4>();
      Ptr<Ipv4RoutingProtocol> p = ipv4->GetRoutingProtocol ();
      Ptr<ns3::gpsr::RoutingProtocol> rp = DynamicCast<ns3::gpsr::RoutingProtocol> (p);
      rp->SetRadius (newRadius);
    }

  // set timeout on scaling (this function is called
  // once per second so it is ok to do this here,
  // not pretty - downright ugly in fact - but it
  // will work so fuck it)
  scaling = SCALING_DURATION;
}

  /**************************************************************************
	                   ..::  MAIN CODE ::..
  ***************************************************************************/

int main(int argc, char *argv[])
{
  std::string phyMode("OfdmRate6MbpsBW10MHz");

  bool verbose = false;
  int simDuration = SIM_DURATION;
  int nodeCount = NODE_COUNT; //3306; // 3306
  int policeCount = POLICE_COUNT;
  int port = 6;

  // survaillance zone
  int x = 2500 * SCALE;
  int y = 2500 * SCALE;
  int r = 500 * SCALE;

  // needed for pyvis
  CommandLine cmd;
  cmd.Parse (argc, argv);

  // Convert to time object
  //double interval = 1.0; // seconds
  //Time interPacketInterval = Seconds(interval);

  /**************************************************************************
	 NODES, WIFI, INTERFACES...
  ***************************************************************************/

  // we need to create two separate classes of nodes, one for emergency
  // services and one for general cars, a and b respectively
  //NodeContainer aNodes;
  NodeContainer nodes;
  nodes.Create(nodeCount);
  // nodes.Create (3360);

  // The below set of helpers will help us to put together the wifi NICs we want
  YansWifiPhyHelper wifiPhy = YansWifiPhyHelper::Default();

  wifiPhy.Set ("TxPowerStart", DoubleValue(33.0));
  wifiPhy.Set ("TxPowerEnd", DoubleValue(33.0));
  wifiPhy.Set ("TxPowerLevels", UintegerValue(1));
  wifiPhy.Set ("TxGain", DoubleValue (0));
  wifiPhy.Set ("RxGain", DoubleValue (0));

  YansWifiChannelHelper wifiChannel = YansWifiChannelHelper::Default();

  wifiChannel.AddPropagationLoss ("ns3::NakagamiPropagationLossModel");

  Ptr < YansWifiChannel > channel = wifiChannel.Create();
  wifiPhy.SetChannel(channel);
  // ns-3 supports generate a pcap trace
  wifiPhy.SetPcapDataLinkType(YansWifiPhyHelper::DLT_IEEE802_11);
  NqosWaveMacHelper wifi80211pMac = NqosWaveMacHelper::Default();
  Wifi80211pHelper wifi80211p = Wifi80211pHelper::Default();
  if (verbose) {
    wifi80211p.EnableLogComponents();    // Turn on all Wifi 802.11p logging
  }
  wifi80211p.SetRemoteStationManager("ns3::ConstantRateWifiManager",
                                     "DataMode", StringValue(phyMode),
                                     "ControlMode",
                                     StringValue(phyMode));
  NetDeviceContainer devices = wifi80211p.Install(wifiPhy, wifi80211pMac,
                                                  nodes);

  // Tracing
  // wifiPhy.EnablePcap ("pcap/wave-simple-80211p", devices);

  // set up GPSR
  GpsrHelper gpsr;
  //gpsr.Set ("Random", m_random);
  InternetStackHelper internet;
  internet.SetRoutingHelper(gpsr);
  internet.Install(nodes);
  Ipv4AddressHelper ipv4;
  NS_LOG_INFO("Assign IP Addresses.");
  ipv4.SetBase("10.0.0.0", "255.255.0.0");
  gpsr.Install ();
  Ipv4InterfaceContainer interfaces = ipv4.Assign(devices);

  /**************************************************************************
	 NETWORK TRAFFIC
  ***************************************************************************/

  // data rate and packet size
  Config::SetDefault  ("ns3::UdpClient::PacketSize",
                       UintegerValue (1024));//419400));
  Config::SetDefault  ("ns3::UdpClient::Interval", TimeValue (Seconds (1.0))); //0.1
  uint64_t max = UINT32_MAX - 1;
  Config::SetDefault  ("ns3::UdpClient::MaxPackets", UintegerValue (max));

  // initialise with clients and servers for police cars
  {
    int i;
    for (i = 0; i < policeCount; i++)
      {
        StreamData *sd = new StreamData;
        sd->streaming = false;
        sd->policeCar = true;
        nodeStreams.push_back (sd);

        // create the server
        UdpServerHelper Server (port);
        serverApps.Add (Server.Install (nodes.Get (i)));

        // set the surveillance zone
        Ptr<Node> n = nodes.Get (i);
        Ptr<Ipv4> ipv4 = n->GetObject<Ipv4>();
        Ptr<Ipv4RoutingProtocol> p = ipv4->GetRoutingProtocol ();
        Ptr<ns3::gpsr::RoutingProtocol> rp = DynamicCast<ns3::gpsr::RoutingProtocol> (p);
        rp->SetZone (x, y, r);
        rp->SetPoliceNode (); // signify police node
      }
    for (; i < nodeCount; i++)
      {
        StreamData *sd = new StreamData;
        sd->streaming = false;
        sd->policeCar = false;
        nodeStreams.push_back (sd);

        // create the client
        for (int j = 0; j < policeCount; j++)
          {
            UdpClientHelper client (interfaces.GetAddress (j), port + 1);
            clientApps.Add(client.Install (nodes.Get (i)));
            // can be on at start since no SREQ
            // make sure client is off at the start
            // Ptr<Application> a = clientApps.Get (clientApps.GetN () - 1);
            // Ptr<UdpClient> c = DynamicCast<UdpClient> (a);
            // stop time cannot be now, needs to be in the future
            // even if only fractionally
            // c->SetStopTime (ASAP);

            // set the surveillance zone
            Ptr<Node> n = nodes.Get (i);
            Ptr<Ipv4> ipv4 = n->GetObject<Ipv4>();
            Ptr<Ipv4RoutingProtocol> p = ipv4->GetRoutingProtocol ();
            Ptr<ns3::gpsr::RoutingProtocol> rp =
              DynamicCast<ns3::gpsr::RoutingProtocol> (p);
            rp->SetZone (x, y, r);
          }
      }
  }

  // every second check the position and status of each node so that streaming
  // can be stopped/started for each node
  for (int i = 0; i < simDuration; i++)
    {
      Simulator::Schedule(Seconds(i), &videoStreamingHandler, policeCount);
    }

  /**************************************************************************
	 MOBILITY
  ***************************************************************************/

  /************* police nodes *************/

  MobilityHelper policeMobility;
  //ObjectFactory mobilityFactory;

  policeMobility.SetMobilityModel ("ns3::WaypointMobilityModel");
  //mobilityFactory.SetTypeId ("ns3::WaypointMobilityModel");

  // create waypoints for the police nodes
  // all have same destinationPort
  Waypoint dst (Seconds (175.0), Vector (2500 * SCALE, 2500 * SCALE, 0));
  Waypoint orn[] = {Waypoint(Seconds (0.0), Vector (2500 * SCALE, 500 * SCALE, 0)),
                    Waypoint(Seconds (0.0), Vector (500 * SCALE, 2500 * SCALE, 0)),
                    Waypoint(Seconds (0.0), Vector (2500 * SCALE, 4500 * SCALE, 0)),
                    Waypoint(Seconds (0.0), Vector (4500 * SCALE, 2500 * SCALE, 0)),
                    Waypoint(Seconds (0.0), Vector (500 * SCALE, 100 * SCALE, 0)),
                    Waypoint(Seconds (0.0), Vector (300 * SCALE, 500 * SCALE, 0)),
                    Waypoint(Seconds (0.0), Vector (3100 * SCALE, 500 * SCALE, 0)),
                    Waypoint(Seconds (0.0), Vector (1000 * SCALE, 4500 * SCALE, 0)),
                    Waypoint(Seconds (0.0), Vector (4000 * SCALE, 4500 * SCALE, 0))};

  for (int i = 0; i < policeCount; i++)
    {
      policeMobility.Install (nodes.Get (i));
      Ptr<WaypointMobilityModel> mm = nodes.Get (i)->
        GetObject<WaypointMobilityModel> ();
      mm->AddWaypoint (orn[i]);
      mm->AddWaypoint (dst);
    }

  /************* regular nodes ************/

  MobilityHelper randomMobility;
  int nodeSpeed = 10; // m/s
  int nodePause = 2;

  // 5km x 5km area
  ObjectFactory pos;
  pos.SetTypeId ("ns3::RandomRectanglePositionAllocator");
  pos.Set ("X", StringValue ("ns3::UniformRandomVariable[Min=0.0|Max=500.0]"));
  pos.Set ("Y", StringValue ("ns3::UniformRandomVariable[Min=0.0|Max=500.0]"));

  Ptr<PositionAllocator> fixedPositionAlloc = pos.Create ()->
    GetObject<PositionAllocator> ();

  std::stringstream ssSpeed;
  ssSpeed << "ns3::UniformRandomVariable[Min=4|Max=" << nodeSpeed << "]";
  std::stringstream ssPause;
  ssPause << "ns3::ConstantRandomVariable[Constant=" << nodePause << "]";
  randomMobility.SetMobilityModel("ns3::RandomWaypointMobilityModel",
                            "Speed", StringValue (ssSpeed.str ()),
                            "Pause", StringValue (ssPause.str ()),
                            "PositionAllocator",
                                  PointerValue (fixedPositionAlloc));
  randomMobility.SetPositionAllocator(fixedPositionAlloc);
  //  randomMobility.Install(nodes);
  // install for non-police nodes
  for (int i = policeCount; i < nodeCount; i++)
    randomMobility.Install (nodes.Get(i));

  //
  //exit(1);

  /**************************************************************************
	 SIMULATOR OPTIONS
  ***************************************************************************/
  AsciiTraceHelper ascii;
  wifiPhy.EnableAsciiAll(ascii.CreateFileStream("tars.tr"));

  // enable output suitable for NetAnim
  AnimationInterface anim ("test_tars.xml");
  anim.SetMaxPktsPerTraceFile (UINT64_MAX);
  anim.EnablePacketMetadata(true);

  // enable flow monitoring
  FlowMonitorHelper flowmonHelper;
  Ptr<FlowMonitor> flowmon = flowmonHelper.InstallAll ();

  Simulator::Stop(Seconds(simDuration));
  Simulator::Run();

  /**************************************************************************
	 DATA COLLECTION
  ***************************************************************************/

  // output flowmon data
  flowmon->CheckForLostPackets ();
  flowmon->SerializeToXmlFile ("FlowMonitorResults.xml", true, true);

  Ptr<Ipv4FlowClassifier> classifier = DynamicCast<Ipv4FlowClassifier> (flowmonHelper.GetClassifier ());
  std::map<FlowId, FlowMonitor::FlowStats> stats = flowmon->GetFlowStats ();

  uint32_t txPacketsum = 0;
  uint32_t rxPacketsum = 0;
  uint32_t DropPacketsum = 0;
  uint32_t LostPacketsum = 0;
  uint32_t rxBytessum = 0;
  double Delaysum = 0;

  for (std::map<FlowId, FlowMonitor::FlowStats>::const_iterator iter = stats.begin (); iter != stats.end (); ++iter)
    {
      //Ipv4FlowClassifier::FiveTuple t = classifier->FindFlow (iter->first);

      //if (t.destinationPort == port)
      //  {
          txPacketsum += iter->second.txPackets;
          rxPacketsum += iter->second.rxPackets;
          LostPacketsum += iter->second.lostPackets;
          DropPacketsum += iter->second.packetsDropped.size();
          Delaysum += iter->second.delaySum.GetSeconds();
          rxBytessum += iter->second.rxBytes;
          // }
    }

  // write the data to file
  std::cout << ((rxPacketsum * 100) / txPacketsum)
            << ","
            << ((LostPacketsum * 100) / txPacketsum)
            << ","
            << ((rxBytessum * 8.0) / SIM_DURATION) / 1024 / POLICE_COUNT
            << ","
            << (Delaysum / rxPacketsum) * 1000
            << std::endl;


  /**************************************************************************
	 Simulator CLEANUP
  ***************************************************************************/

  Simulator::Destroy();

  /**************************************************************************
	 POST SIMULATION ANALYSIS
  ***************************************************************************/

  // write extra data
  std::ofstream outfile ("time-radius-nodesin.txt");
  outfile << "time,radius,members" << std::endl;
  for (unsigned int i = 0; i < radii.size(); i++)
    {
      outfile << i << "," << radii[i] << "," << memberCount[i]
              << "," << std::endl;
    }

  outfile.close();

  return 0;
}
